@@ -5,19 +5,18 @@ from __future__ import annotations
 # Will no longer be needed in Python 3.10.
 
 import abc
-import asyncio
 import concurrent.futures
 import contextlib
 import datetime
-import functools
 import logging
 import os
 import os.path
 import pathlib
+import sys
+import time
 from dataclasses import dataclass
-from functools import partial
 from io import UnsupportedOperation
-from typing import List, Iterator, TypeVar, Any, Optional, Callable, AsyncIterator
+from typing import List, Iterator, Tuple, Type, TypeVar, Any, Optional, Union, Callable
 
 from .serializer import (
     ByteSerializer, TextSerializer,
@@ -44,26 +43,76 @@ class FileInfo:
     details: Any   # platform-dependent
 
 
-def _execute_in_thread_pool(jobs, concurrency: int = None):
+def _execute_in_thread_pool(
+        jobs,
+        concurrency: int = None,
+        *,
+        retry_on_exceptions: Optional[Union[Type[Exception],
+                                            Tuple[Type[Exception]]]] = None,
+        retries: int = 3,
+):
+    '''
+    If you want to skip certain errors and don't retry on them,
+    use `retries=0`.
+    '''
     if concurrency is None:
-        concurrency = 4
+        concurrency = 16
     else:
-        assert 0 <= concurrency <= 32
+        assert 0 <= concurrency <= 64
+        if concurrency < 1:
+            concurrency = 1
 
-    if concurrency <= 1:
+    class MyExc(Exception):
+        pass
+
+    if retry_on_exceptions is None:
+        def ff(f, args, kwargs):
+            return f(*args, **kwargs)
+    else:
+        def ff(f, args, kwargs):
+            try:
+                return f(*args, **kwargs)
+            except retry_on_exceptions:
+                return MyExc(f, args, kwargs)
+
+    with concurrent.futures.ThreadPoolExecutor(concurrency) as pool:
         results = []
-        for f, args, kwargs in jobs:
-            results.append(f(*args, **kwargs))
+        tasks = []
+        bad = []
+        for args in jobs:
+            tasks.append(pool.submit(ff, *args))
+        for f in concurrent.futures.as_completed(tasks):
+            z = f.result()
+            if isinstance(z, MyExc):
+                bad.append(z.args)
+            else:
+                results.append(z)
+        jobs = bad
+
+    if not jobs:
         return results
 
-    pool = concurrent.futures.ThreadPoolExecutor(concurrency)
-    tasks = []
-    for f, args, kwargs in jobs:
-        tasks.append(pool.submit(f, *args, **kwargs))
-    results = []
-    for f in concurrent.futures.as_completed(tasks):
-        results.append(f.result())
-    return results
+    # Switch to sequential retry.
+    # Some issues could be due to large files.
+    for _ in range(retries):
+        logger.warning('failed on %d items; retrying', len(jobs))
+        time.sleep(0.6)
+        bad = []
+        for args in jobs:
+            z = ff(*args)
+            if isinstance(z, MyExc):
+                bad.append(z.args)
+            else:
+                results.append(z)
+        if not bad:
+            break
+        jobs = bad
+
+    if not jobs:
+        return results
+
+    print(f'failed on {len(jobs)} items: {jobs}', file=sys.stderr)
+    raise RuntimeError(f'failed on {len(jobs)} items')
 
 
 def _should_update(source: Upath, target: Upath) -> bool:
@@ -107,56 +156,30 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         return True
 
     @classmethod
-    def register_read_write_byte_format(cls, serde: ByteSerializer, name: str):
+    def register_read_write_byte_format(cls, serde: Type[ByteSerializer], name: str):
         def _write(self, data, *, overwrite: bool = False):
             return self.write_bytes(serde.serialize(data), overwrite=overwrite)
 
-        async def _a_write(self, data, *, overwrite: bool = False):
-            return await asyncio.get_running_loop().run_in_executor(
-                None,
-                functools.partial(_write, overwrite=overwrite), self, data,
-            )
-
         def _read(self):
             z = self.read_bytes()
             return serde.deserialize(z)
 
-        async def _a_read(self):
-            return await asyncio.get_running_loop().run_in_executor(
-                None, _read, self,
-            )
-
         setattr(cls, f'write_{name}', _write)
-        setattr(cls, f'a_write_{name}', _a_write)
         setattr(cls, f'read_{name}', _read)
-        setattr(cls, f'a_read_{name}', _a_read)
 
     @classmethod
-    def register_read_write_text_format(cls, serde: TextSerializer, name: str):
+    def register_read_write_text_format(cls, serde: Type[TextSerializer], name: str):
         def _write(self, data, *, overwrite: bool = False):
             return self.write_text(serde.serialize(data), overwrite=overwrite)
 
-        async def _a_write(self, data, *, overwrite: bool = False):
-            return await asyncio.get_running_loop().run_in_executor(
-                None,
-                functools.partial(_write, overwrite=overwrite), self, data,
-            )
-
         def _read(self):
             z = self.read_text()
             return serde.deserialize(z)
 
-        async def _a_read(self):
-            return await asyncio.get_running_loop().run_in_executor(
-                None, _read, self,
-            )
-
         setattr(cls, f'write_{name}', _write)
-        setattr(cls, f'a_write_{name}', _a_write)
         setattr(cls, f'read_{name}', _read)
-        setattr(cls, f'a_read_{name}', _a_read)
 
-    def __init__(self, *pathsegments: str, **kwargs):
+    def __init__(self, *pathsegments: str):
         '''`Upath` is the base class for a client to a blob store,
         including local file system as a special case.
 
@@ -168,13 +191,6 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         For a local POSIX file system, the root is the regular `/`.
         For Azure blob store, the root is that of a "container".
         For AWS and GCP blob stores, the root is that of a "bucket".
-
-        `**kwargs`: additional arguments. This usually concerns
-        credentials and top-level "divisions" for a blob store.
-        (For example, "container" for Azure, "bucket" for AWS or GCP.)
-        In general, these arguments should be read-only
-        and remain unchanged during the lifetime of the object.
-        For local file system, this is not needed (refer to class `LocalUpath`).
         '''
 
         self._path = os.path.normpath(os.path.join(
@@ -182,15 +198,6 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         # The path is always "absolute" starting with '/'.
         # Unless it is `/`, it does not have a trailing `/`.
 
-        self._kwargs = kwargs
-        # The extra `kwargs` is handled this way so that
-        # the few functions that calls `self.__class__(...)`
-        # work with subclasses that take additional arguments
-        # in their `__init__`.
-        # A subclass that uses such extra arguments may need to
-        # redefine the "comparison" special methods to take
-        # into account some of these parameters as needed.
-
     def __repr__(self) -> str:
         return f"{self.__class__.__name__}('{self._path}')"
 
@@ -234,6 +241,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                  concurrency: int = None,
                  exist_action: str = None,
                  update_filter: Callable[[Upath, Upath], bool] = None,
+                 **kwargs,
                  ) -> int:
         '''Analogous to `copy_file`.
 
@@ -252,7 +260,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                     {'exist_action': exist_action, 'update_filter': update_filter},
                 )
 
-        nn = _execute_in_thread_pool(foo(), concurrency)
+        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
         return sum(nn)
 
     def _copy_file(self: T, target: T) -> None:
@@ -335,6 +343,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                    concurrency: int = None,
                    exist_action: str = None,
                    update_filter: Callable[[Upath, Upath], bool] = None,
+                   **kwargs,
                    ) -> int:
         '''Copy the content of `self` to the specified `target`,
         which is typically in another store.
@@ -351,15 +360,18 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         Return the number of files copied.
         '''
         def foo():
+            self_path = self.path
             for p in self.riterdir():
-                extra = str(p.path.relative_to(self.path))
+                extra = str(p.path.relative_to(self_path))
                 yield (
                     p.export_file,
                     [target / extra],
-                    {'exist_action': exist_action, 'update_filter': update_filter},
+                    {'exist_action': exist_action,
+                     'check_source_exist': False,
+                     'update_filter': update_filter},
                 )
 
-        nn = _execute_in_thread_pool(foo(), concurrency)
+        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
         return sum(nn)
 
     def _export_file(self, target: Upath) -> None:
@@ -374,6 +386,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                     *,
                     exist_action: str = None,
                     update_filter: Callable[[Upath, Upath], bool] = None,
+                    check_source_exist: bool = True,
                     ) -> int:
         '''Copy the file to the specified `target`, which is typically
         in another store.
@@ -387,7 +400,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
 
         Compare with `copy_file`, which make copies within the same store.
         '''
-        if not self.is_file():
+        if check_source_exist and not self.is_file():
             raise FileNotFoundError(self)
 
         if target.is_file():
@@ -395,10 +408,8 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                                       exist_action=exist_action,
                                       update_filter=update_filter):
                 target.remove_file()
-                logger.info("copying '%s' to '%s'", self, target)
-                self._export_file(target)
-                return 1
-            return 0
+            else:
+                return 0
 
         if target.is_dir():
             # Do not delete.
@@ -421,13 +432,24 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                    concurrency: int = None,
                    exist_action: str = None,
                    update_filter: Callable[[Upath, Upath], bool] = None,
+                   **kwargs,
                    ) -> int:
         '''Analogous to `export_dir`.
         '''
-        return source.export_dir(self,
-                                 concurrency=concurrency,
-                                 exist_action=exist_action,
-                                 update_filter=update_filter)
+        def foo():
+            source_path = source.path
+            for p in source.riterdir():
+                extra = str(p.path.relative_to(source_path))
+                yield (
+                    (self / extra).import_file,
+                    [p],
+                    {'exist_action': exist_action,
+                     'check_source_exist': False,
+                     'update_filter': update_filter},
+                )
+
+        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
+        return sum(nn)
 
     def _import_file(self, source: Upath) -> None:
         # `self` does not exist, hence no concern about overwriting.
@@ -441,11 +463,26 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
     def import_file(self, source: Upath, *,
                     exist_action: str = None,
                     update_filter: Callable[[Upath, Upath], bool] = None,
+                    check_source_exist: bool = True,
                     ) -> int:
-        return source.export_file(self,
-                                  exist_action=exist_action,
-                                  update_filter=update_filter,
-                                  )
+        if check_source_exist and not source.is_file():
+            raise FileNotFoundError(source)
+
+        if self.is_file():
+            if self._should_overwrite(source, self,
+                                      exist_action=exist_action,
+                                      update_filter=update_filter):
+                self.remove_file()
+            else:
+                return 0
+
+        if self.is_dir():
+            # Do not delete.
+            raise FileExistsError(self)
+
+        logger.info("copying '%s' to '%s'", source, self)
+        self._import_file(source)
+        return 1
 
     @ abc.abstractmethod
     def is_dir(self) -> bool:
@@ -507,7 +544,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
 
     def joinpath(self: T, *other: str) -> T:
         '''Join this path with more segments, return the new path object.'''
-        return self.__class__(self._path, *other, **self._kwargs)
+        return self.with_path(self._path, *other)
 
     @ contextlib.contextmanager
     @ abc.abstractmethod
@@ -558,7 +595,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
 
     @ property
     def parent(self: T) -> T:
-        return self.__class__(str(self.path.parent), **self._kwargs)
+        return self.with_path(str(self.path.parent))
 
     @ property
     def path(self) -> pathlib.PurePosixPath:
@@ -589,7 +626,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         # Refer to https://docs.python.org/3/library/functions.html#open
         return self.read_bytes().decode(encoding=encoding, errors=errors)
 
-    def remove_dir(self, *, concurrency: int = None) -> int:
+    def remove_dir(self, *, concurrency: int = None, **kwargs) -> int:
         '''Remove the directory pointed to by `self`,
         along with all its contents, recursively.
 
@@ -605,7 +642,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
             for p in self.riterdir():
                 yield p.remove_file, [], {}
 
-        nn = _execute_in_thread_pool(foo(), concurrency)
+        nn = _execute_in_thread_pool(foo(), concurrency, **kwargs)
         return sum(nn)
 
     @abc.abstractmethod
@@ -623,6 +660,8 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                    target: str,
                    *,
                    concurrency: int = None,
+                   exist_ok: bool = False,
+                   **kwargs,
                    ) -> T:
         '''Analogous to `rename_file`.
 
@@ -636,8 +675,9 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         if target_ == self:
             return self
 
-        if target_.exists():
-            raise FileExistsError(target_)
+        if not exist_ok:
+            if target_.exists():
+                raise FileExistsError(target_)
 
         def foo():
             for p in self.riterdir():
@@ -648,7 +688,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                     {},
                 )
 
-        _ = _execute_in_thread_pool(foo(), concurrency)
+        _ = _execute_in_thread_pool(foo(), concurrency, **kwargs)
         return target_
 
     def _rename_file(self: T, target: T):
@@ -730,18 +770,30 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         return self.path.suffixes
 
     def with_name(self: T, name: str) -> T:
-        return self.__class__(str(self.path.with_name(name)), **self._kwargs)
+        return self.with_path(str(self.path.with_name(name)))
+
+    def with_path(self: T, *paths) -> T:
+        '''
+        Returns a new object of the same class at the specified path.
+        The new path is unrelated to the current path; in other words,
+        the new path is not "relative" to the current path.
+
+        Meta data such as account info remains the same.
+
+        Subclass needs to reimplement this method
+        if its `__init__` expects additional args.
+        '''
+        return self.__class__(*paths)
 
     # def with_stem(self: T, stem: str) -> T:
     #     # Available in Python 3.9+.
-    #     return self.__class__(str(self.path.with_stem(stem)), **self._kwargs)
+    #     return self.with_path(str(self.path.with_stem(stem)))
 
     def with_suffix(self: T, suffix: str) -> T:
         '''`suffix` should include a dot, like '.txt'.
         If `suffix` is '', the effect is to remove the existing suffix.
         '''
-        return self.__class__(str(self.path.with_suffix(suffix)),
-                              **self._kwargs)
+        return self.with_path(str(self.path.with_suffix(suffix)))
 
     @ abc.abstractmethod
     def write_bytes(self,
@@ -785,59 +837,10 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         self.write_bytes(z, overwrite=overwrite)
         return n
 
-    async def a_iterdir(self: T) -> AsyncIterator[T]:
-        # TODO: may need reimplementation.
-        for p in self.iterdir():
-            yield p
 
-    @ contextlib.asynccontextmanager
-    async def a_lock(self, *, timeout: int = None):
-        # TODO: a naive implementation.
-        with self.lock(timeout=timeout):
-            yield
-
-    async def a_riterdir(self: T) -> AsyncIterator[T]:
-        # TODO: may need reimplementation.
-        for p in self.riterdir():
-            yield p
-
-
-# Add methods 'read_json', 'write_json', 'a_read_json', 'a_write_json', etc.
+# Add methods 'read_json', 'write_json', 'read_pickle', 'write_pickle', etc.
 Upath.register_read_write_text_format(JsonSerializer, 'json')
 Upath.register_read_write_byte_format(PickleSerializer, 'pickle')
 Upath.register_read_write_byte_format(CompressedPickleSerializer, 'pickle_z')
 Upath.register_read_write_byte_format(OrjsonSerializer, 'orjson')
 Upath.register_read_write_byte_format(CompressedOrjsonSerializer, 'orjson_z')
-
-
-def make_a_method(name):
-    '''
-    Create an async method named f'a_{name}' based on
-    the sync method 'name'. The async method has the same
-    interface as the sync one, i.e. they take the same parameters.
-    '''
-    async def f(self, *args, **kwargs):
-        f = partial(getattr(self, name), *args, **kwargs)
-        return await asyncio.get_running_loop().run_in_executor(
-            None, f
-        )
-
-    f.__name__ = f'a_{name}'
-    return f
-
-
-# Add async methods 'a_copy_dir', 'a_copy_file', etc.
-for m in ('copy_dir', 'copy_file',
-          'export_dir', 'export_file',
-          'exists',
-          'file_info',
-          'import_dir', 'import_file',
-          'is_dir', 'is_file',
-          'ls',
-          'read_bytes', 'read_text',
-          'remove_dir', 'remove_file',
-          'rename_dir', 'rename_file',
-          'rmrf',
-          'write_bytes', 'write_text',
-          ):
-    setattr(Upath, f'a_{m}', make_a_method(m))
